""":mod:`libearth.parser.atom` --- Atom parser
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Parsing Atom feed. Atom specification is :rfc:`4287`

.. todo::

   Parsing text construct which ``type`` is ``'xhtml'``.

"""
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from ..codecs import Rfc3339, Rfc822
from ..compat.etree import fromstring
from ..feed import (Category, Content, Entry, Feed, Generator, Link,
                    Person, Source, Text)
from ..schema import DecodeError
from .util import normalize_xml_encoding

__all__ = 'ATOM_XMLNS_SET', 'XML_XMLNS', 'parse_atom'


#: (:class:`frozenset`) The set of XML namespaces for Atom format.
ATOM_XMLNS_SET = frozenset([
    'http://www.w3.org/2005/Atom',
    'http://purl.org/atom/ns#'
])

#: (:class:`str`) The XML namespace for the predefined ``xml:`` prefix.
XML_XMLNS = 'http://www.w3.org/XML/1998/namespace'


def parse_atom(xml, feed_url, parse_entry=True):
    """Atom parser.  It parses the Atom XML and returns the feed data
    as internal representation.

    :param xml: target atom xml to parse
    :type xml: :class:`str`
    :param feed_url: the url used to retrieve the atom feed.
                     it will be the base url when there are any relative
                     urls without ``xml:base`` attribute
    :type feed_url: :class:`str`
    :param parse_entry: whether to parse inner items as well.
                        it's useful to ignore items when retrieve
                        ``<source>`` in rss 2.0.  :const:`True` by default.
    :type parse_entry: :class:`bool`
    :returns: a pair of (:class:`~libearth.feed.Feed`, crawler hint)
    :rtype: :class:`tuple`

    """
    root = fromstring(normalize_xml_encoding(xml))
    for atom_xmlns in ATOM_XMLNS_SET:
        if root.tag.startswith('{' + atom_xmlns + '}'):
            break
    feed_data = atom_get_feed_data(root, feed_url, atom_xmlns)
    entries = root.findall('{' + atom_xmlns + '}entry')
    if parse_entry:
        entries_data = atom_get_entry_data(entries, feed_url, atom_xmlns)
        feed_data.entries = entries_data
    return feed_data, None


def atom_parse_text_construct(data):
    text = Text()
    text_type = data.get('type')
    if text_type is not None:
        if text_type == 'text/plain':
            text_type = 'text'
        elif text_type == 'text/html':
            text_type = 'html'
        text.type = text_type
    if text.type in ('text', 'html'):
        text.value = data.text
    elif text.value == 'xhtml':
        text.value = ''  # TODO
    return text


def atom_parse_person_construct(data, xml_base, atom_xmlns, as_list=False):
    person = Person()
    xml_base = atom_get_xml_base(data, xml_base)
    for child in data:
        if child.tag == '{' + atom_xmlns + '}name':
            person.name = child.text
        elif child.tag == '{' + atom_xmlns + '}uri':
            person.uri = urlparse.urljoin(xml_base, child.text)
        elif child.tag == '{' + atom_xmlns + '}email':
            person.email = child.text
    if not person.name:
        if person.email:
            person.name = person.email
        elif person.uri:
            person.name = person.uri
        else:
            person = None
    if as_list:
        return [person] if person else []
    return person


def atom_get_feed_data(root, feed_url, atom_xmlns):
    feed_data = Feed()
    xml_base = atom_get_xml_base(root, feed_url)
    alt_id = None
    for data in root:
        if data.tag == '{' + atom_xmlns + '}id':
            feed_data.id = alt_id = atom_get_id_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}title':
            feed_data.title = atom_get_title_tag(data)
        elif data.tag == '{' + atom_xmlns + '}updated':
            feed_data.updated_at = atom_parse_datetime(data)
        elif data.tag == '{' + atom_xmlns + '}modified':
            # Non-standard: some feeds use <modified> instead of <updated>
            # which is standard e.g. Naver Blog.
            if not feed_data.updated_at:
                feed_data.updated_at = atom_parse_datetime(data)
        elif data.tag == '{' + atom_xmlns + '}author':
            feed_data.authors.extend(
                atom_parse_person_construct(data, xml_base, atom_xmlns, True)
            )
        elif data.tag == '{' + atom_xmlns + '}category':
            category = atom_get_category_tag(data)
            if category:
                feed_data.categories.append(atom_get_category_tag(data))
        elif data.tag == '{' + atom_xmlns + '}contributor':
            feed_data.contributors.extend(
                atom_parse_person_construct(data, xml_base, atom_xmlns, True)
            )
        elif data.tag == '{' + atom_xmlns + '}link':
            link = atom_get_link_tag(data, xml_base)
            if link.relation == 'self':
                alt_id = alt_id or link.uri
            feed_data.links.append(link)
        elif data.tag == '{' + atom_xmlns + '}generator':
            feed_data.generator = atom_get_generator_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}icon':
            feed_data.icon = atom_get_icon_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}logo':
            feed_data.logo = atom_get_logo_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}rights':
            feed_data.rights = atom_get_rights_tag(data)
        elif data.tag == '{' + atom_xmlns + '}subtitle':
            feed_data.subtitle = atom_get_subtitle_tag(data)
        elif data.tag == '{' + atom_xmlns + '}entry':
            break
    if feed_data.id is None:
        feed_data.id = alt_id or feed_url
    return feed_data


def atom_get_entry_data(entries, feed_url, atom_xmlns):
    entries_data = []
    for entry in entries:
        entry_data = Entry()
        xml_base = atom_get_xml_base(entry, feed_url)
        for data in entry:
            if data.tag == '{' + atom_xmlns + '}id':
                entry_data.id = atom_get_id_tag(data, xml_base)
            elif data.tag == '{' + atom_xmlns + '}title':
                entry_data.title = atom_get_title_tag(data)
            elif data.tag == '{' + atom_xmlns + '}updated':
                entry_data.updated_at = atom_parse_datetime(data)
            elif data.tag == '{' + atom_xmlns + '}modified':
                # Non-standard: some feeds use <modified> instead of <updated>
                # which is standard e.g. Naver Blog.
                if not entry_data.updated_at:
                    entry_data.updated_at = atom_parse_datetime(data)
            elif data.tag == '{' + atom_xmlns + '}author':
                entry_data.authors.extend(
                    atom_parse_person_construct(
                        data, xml_base, atom_xmlns,
                        as_list=True
                    )
                )
            elif data.tag == '{' + atom_xmlns + '}category':
                category = atom_get_category_tag(data)
                if category:
                    entry_data.categories.append(atom_get_category_tag(data))
            elif data.tag == '{' + atom_xmlns + '}contributor':
                entry_data.contributors.extend(
                    atom_parse_person_construct(
                        data, xml_base, atom_xmlns,
                        as_list=True
                    )
                )
            elif data.tag == '{' + atom_xmlns + '}link':
                entry_data.links.append(atom_get_link_tag(data, xml_base))
            elif data.tag == '{' + atom_xmlns + '}content':
                entry_data.content = atom_get_content_tag(data, xml_base)
            elif data.tag == '{' + atom_xmlns + '}published':
                entry_data.published_at = atom_parse_datetime(data)
            elif data.tag == '{' + atom_xmlns + '}rights':
                entry_data.rigthts = atom_get_rights_tag(data)
            elif data.tag == '{' + atom_xmlns + '}source':
                entry_data.source = atom_get_source_tag(
                    data, xml_base, atom_xmlns
                )
            elif data.tag == '{' + atom_xmlns + '}summary':
                entry_data.summary = atom_get_summary_tag(data)
        entries_data.append(entry_data)
    return entries_data


def atom_get_xml_base(data, default):
    if '{' + XML_XMLNS + '}base' in data.attrib:
        return data.attrib['{' + XML_XMLNS + '}base']
    else:
        return default


def atom_get_id_tag(data, xml_base):
    xml_base = atom_get_xml_base(data, xml_base)
    return urlparse.urljoin(xml_base, data.text)


def atom_get_title_tag(data):
    return atom_parse_text_construct(data)


def atom_parse_datetime(data):
    try:
        return Rfc3339().decode(data.text)
    except DecodeError:
        return Rfc822().decode(data.text)


def atom_get_category_tag(data):
    if not data.get('term'):
        return
    category = Category()
    category.term = data.get('term')
    category.scheme_uri = data.get('scheme')
    category.label = data.get('label')
    return category


def atom_get_link_tag(data, xml_base):
    xml_base = atom_get_xml_base(data, xml_base)
    link = Link(
        uri=urlparse.urljoin(xml_base, data.get('href')),
        mimetype=data.get('type'),
        language=data.get('hreflang'),
        title=data.get('title'),
        byte_size=data.get('length')
    )
    rel = data.get('rel')
    if rel:
        link.relation = rel
    return link


def atom_get_generator_tag(data, xml_base):
    generator = Generator()
    xml_base = atom_get_xml_base(data, xml_base)
    generator.value = data.text
    if 'uri' in data.attrib:
        generator.uri = urlparse.urljoin(xml_base, data.attrib['uri'])
    generator.version = data.get('version')
    return generator


def atom_get_icon_tag(data, xml_base):
    xml_base = atom_get_xml_base(data, xml_base)
    return urlparse.urljoin(xml_base, data.text)


def atom_get_logo_tag(data, xml_base):
    xml_base = atom_get_xml_base(data, xml_base)
    return urlparse.urljoin(xml_base, data.text)


def atom_get_rights_tag(data):
    return atom_parse_text_construct(data)


def atom_get_subtitle_tag(data):
    return atom_parse_text_construct(data)


def atom_get_content_tag(data, xml_base):
    content = Content()
    content.value = data.text
    content_type = data.get('type')
    if content_type is not None:
        content.type = content_type
    if 'src' in data.attrib:
        content.source_uri = urlparse.urljoin(xml_base, data.attrib['src'])
    return content


def atom_get_source_tag(data_dump, xml_base, atom_xmlns):
    source = Source()
    xml_base = atom_get_xml_base(data_dump[0], xml_base)
    authors = []
    categories = []
    contributors = []
    links = []
    for data in data_dump:
        xml_base = atom_get_xml_base(data, xml_base)
        if data.tag == '{' + atom_xmlns + '}author':
            authors.extend(
                atom_parse_person_construct(data, xml_base, atom_xmlns, True)
            )
            source.authors = authors
        elif data.tag == '{' + atom_xmlns + '}category':
            categories.append(atom_get_category_tag(data))
            source.categories = categories
        elif data.tag == '{' + atom_xmlns + '}contributor':
            contributors.extend(
                atom_parse_person_construct(data, xml_base, atom_xmlns, True)
            )
            source.contributors = contributors
        elif data.tag == '{' + atom_xmlns + '}link':
            links.append(atom_get_link_tag(data, xml_base))
            source.links = links
        elif data.tag == '{' + atom_xmlns + '}id':
            source.id = atom_get_id_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}title':
            source.title = atom_get_title_tag(data)
        elif data.tag == '{' + atom_xmlns + '}updated':
            source.updated_at = atom_parse_datetime(data)
        elif data.tag == '{' + atom_xmlns + '}generator':
            source.generator = atom_get_generator_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}icon':
            source.icon = atom_get_icon_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}logo':
            source.logo = atom_get_logo_tag(data, xml_base)
        elif data.tag == '{' + atom_xmlns + '}rights':
            source.rights = atom_get_rights_tag(data)
        elif data.tag == '{' + atom_xmlns + '}subtitle':
            source.subtitle = atom_get_subtitle_tag(data)
    return source


def atom_get_summary_tag(data):
    summary_tag = atom_parse_text_construct(data)
    return summary_tag



""":mod:`libearth.parser.util` --- Utilities for feed parsing
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.3.0

"""
import logging
import re

from ..compat import IRON_PYTHON, binary_type, text_type

__all__ = 'XML_ENCODING_PATTERN ', 'normalize_xml_encoding'


XML_ENCODING_PATTERN = re.compile(br'''
    ^ \s*
    <\?xml
    (?: \s+ version=(?: "[^"]*" | '[^']*')
    |   \s+ encoding=(?: "( [^"]* ) " | '( [^']* )')
    )+
    \s* \?>
''', re.VERBOSE)


def normalize_xml_encoding(document):
    """Normalize the given XML document's encoding to UTF-8 to workaround
    :mod:`xml.etree.ElementTree` module's `encoding detection bug`__.

    __ http://bugs.python.org/issue13612

    .. versionadded:: 0.3.0

    """
    if isinstance(document, text_type):
        return document
    elif not isinstance(document, binary_type):
        raise TypeError('document must be a bytestring or a (unicode) string')
    match = XML_ENCODING_PATTERN.match(document)
    if match:
        encoding = match.group(1) or match.group(2)
        if encoding:
            if not isinstance(encoding, str):
                encoding = encoding.decode('ascii')
            document = document[match.end():]
            try:
                document = document.decode(encoding).encode('utf-8')
            except (LookupError, UnicodeError) as e:
                logger = logging.getLogger(__name__ +
                                           '.normalize_xml_encoding')
                logger.warning(e, exc_info=True)
            if IRON_PYTHON:
                document = bytes(document)
    return document.replace(b'\x1c', b'')



""":mod:`libearth.tz` --- Basic timezone implementations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Almost of this module is from the official documentation of
:mod:`datetime` module in Python standard library.

.. data:: utc

   (:class:`Utc`, :class:`datetime.timezone`) The :class:`~datetime.tzinfo`
   instance that represents UTC.  It's an instance of :class:`Utc`
   in Python 2 (which provide no built-in fixed-offset
   :class:`~datetime.tzinfo` implementation), and an instance of
   :class:`~datetime.timezone` with zero offset in Python 3.

"""
import datetime

from .compat import string_type

__all__ = 'FixedOffset', 'Utc', 'guess_tzinfo_by_locale', 'now', 'utc'


class Utc(datetime.tzinfo):
    """UTC.

    In most cases, it doesn't need to be directly instantiated:
    there's already the :const:`utc` value.

    """

    def __init__(self):
        self.zero = datetime.timedelta(0)

    def utcoffset(self, dt):
        return self.zero

    def dst(self, dt):
        return self.zero

    def tzname(self, dt):
        return 'UTC'

    def __repr__(self):
        cls = type(self)
        return '{0.__module__}.{0.__name__}()'.format(cls)


class FixedOffset(datetime.tzinfo):
    """Fixed offset in minutes east from UTC.

    >>> kst = FixedOffset(9 * 60, name='Asia/Seoul')  # KST +09:00
    >>> current = now()
    >>> current
    datetime.datetime(2013, 8, 15, 3, 18, 37, 404562, tzinfo=libearth.tz.Utc())
    >>> current.astimezone(kst)
    datetime.datetime(2013, 8, 15, 12, 18, 37, 404562,
                      tzinfo=<libearth.tz.FixedOffset Asia/Seoul>)

    """

    def __init__(self, offset, name=None):
        self.offset = datetime.timedelta(minutes=offset)
        self.dst_ = datetime.timedelta(0)
        self.name = name or '{h:+03d}:{m:02d}'.format(h=offset // 60,
                                                      m=offset % 60)

    def utcoffset(self, dt):
        return self.offset

    def dst(self, dt):
        return self.dst_

    def tzname(self, dt):
        return self.name

    def __repr__(self):
        cls = type(self)
        return '<{0.__module__}.{0.__name__} {1}>'.format(cls, self.name)


try:
    utc = datetime.timezone.utc
except AttributeError:
    utc = Utc()


def now():
    """Return the current :class:`~datetime.datetime` with the proper
    :class:`~datetime.tzinfo` setting.

    >>> now()
    datetime.datetime(2013, 8, 15, 3, 17, 11, 892272, tzinfo=libearth.tz.Utc())
    >>> now()
    datetime.datetime(2013, 8, 15, 3, 17, 17, 532483, tzinfo=libearth.tz.Utc())

    """
    return datetime.datetime.utcnow().replace(tzinfo=utc)


LANGUAGE_COUNTRY_TZINFO_TABLE = {
    'ko': {'kr': FixedOffset(9 * 60, 'Asia/Seoul')},
    'ja': {'jp': FixedOffset(9 * 60, 'Asia/Tokyo')}
}


def guess_tzinfo_by_locale(language, country=None):
    """Guess the most commonly used time zone from the given locale.

    :param language: the language code e.g. ``ko``, ``JA``
    :type language: :class:`str`
    :param country: optional country code e.g. ``kr``, ``JP``
    :type country: :class:`str`
    :return: the most commonly used time zone, or :const:`None` if can't
             guess
    :rtype: :class:`datetime.tzinfo`

    .. versionadded:: 0.3.0

    """
    if not isinstance(language, string_type):
        raise TypeError('language must be a string, not ' + repr(language))
    elif not (country is None or isinstance(country, string_type)):
        raise TypeError('country must be a string, not ' + repr(country))
    language = language.strip().lower()
    if len(language) != 2:
        raise ValueError(repr(language) + ' is not a valid language code')
    if country:
        country = country.strip().lower()
        if len(country) != 2:
            raise ValueError(repr(country) + ' is not a valid country code')
    try:
        countries = LANGUAGE_COUNTRY_TZINFO_TABLE[language]
    except KeyError:
        return
    if country:
        return countries.get(country)
    elif len(countries) == 1:
        for tz in countries.values():
            return tz



""":mod:`libearth.codecs` --- Common codecs
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides commonly used codecs to parse RSS-related standard
formats.

"""
import collections
import datetime
import numbers
import platform
import re

from .compat import IRON_PYTHON, string_type
from .schema import Codec, DecodeError, EncodeError
from .tz import FixedOffset, utc

__all__ = 'Boolean', 'Enum', 'Integer', 'Rfc3339', 'Rfc822'


class Enum(Codec):
    """Codec that accepts only predefined fixed types of values::

        gender = Enum(['male', 'female'])

    Actually it doesn't any encoding nor decoding, but it simply *validates*
    all values from XML and Python both.

    Note that values have to consist of only strings.

    :param values: any iterable that yields all possible values
    :type values: :class:`collections.Iterable`

    """

    def __init__(self, values):
        if not isinstance(values, collections.Iterable):
            raise TypeError('enum values must be iterable, not ' +
                            repr(values))
        values = frozenset(values)
        for v in values:
            if not isinstance(v, string_type):
                raise TypeError('enum values must be strings, not ' +
                                repr(v))
        self.values = values

    def encode(self, value):
        if not isinstance(value, string_type):
            raise EncodeError('expected a string, not ' + repr(value))
        elif value not in self.values:
            raise EncodeError(
                '{0!r} is an invalid value; choose one of {1}'.format(
                    value,
                    ', '.join(repr(v) for v in self.values)
                )
            )
        return value

    def decode(self, text):
        if text not in self.values:
            raise DecodeError('expected one of ' +
                              ', '.join(repr(v) for v in self.values))
        return text


class Rfc3339(Codec):
    """Codec to store :class:`datetime.datetime` values to :rfc:`3339`
    format.

    :param prefer_utc: normalize all timezones to UTC.
                       :const:`False` by default
    :type prefer_utc: :class:`bool`

    """

    #: (:class:`re.RegexObject`) The regular expression pattern that
    #: matches to valid :rfc:`3339` date time string.
    PATTERN = re.compile(r'''
        ^
        (?P<year> \d{4} ) - (?P<month> 0[1-9] | 1[012] )
                          - (?P<day> 0[1-9] | [12]\d | 3[01] )
        T
        (?P<hour> [01]\d | 2[0-3] ) : (?P<minute> [0-5]\d )
                                    : (?P<second> [0-5]\d | 60 )
                                    (?: \. (?P<microsecond> \d+ ) )?
        (?P<tz> Z
              | (?P<tz_offset> (?P<tz_offset_sign> [+-] )
                               (?P<tz_offset_hour> [01]\d | 2[0-3] ) :
                               (?P<tz_offset_minute> [0-5]\d ) ) )
        $
    ''', re.VERBOSE)

    def __init__(self, prefer_utc=False):
        self.prefer_utc = bool(prefer_utc)

    def encode(self, value):
        if not isinstance(value, datetime.datetime):
            raise EncodeError(
                '{0.__module__}.{0.__name__} accepts only datetime.datetime '
                'value, not {1!r}'.format(type(self), value)
            )
        elif value.tzinfo is None:
            raise EncodeError(
                '{0.__module__}.{0.__name__} does not accept naive datetime.'
                'datetime value, but {1!r} lacks tzinfo attribute'.format(
                    type(self), value
                )
            )
        if self.prefer_utc:
            value = value.astimezone(utc)
        dt = value.strftime('%Y-%m-%dT%H:%M:%S')
        if value.microsecond:
            # IronPython strftime() seems to ignore %f
            dt += '.{0:06}'.format(value.microsecond).rstrip('0')
        offset = value.tzinfo.utcoffset(value)
        if offset == datetime.timedelta(0):
            dt += 'Z'
        else:
            minutes = offset.seconds // 60
            dt += '{h:+03d}:{m:02d}'.format(h=minutes // 60,
                                            m=minutes % 60)
        return dt

    def decode(self, text):
        text = text.strip()
        match = self.PATTERN.match(text)
        if not match:
            raise DecodeError(repr(text) +
                              ' is not valid RFC 3339 date time string')
        if match.group('tz_offset'):
            tz_hour = int(match.group('tz_offset_hour'))
            tz_minute = int(match.group('tz_offset_minute'))
            tz_sign = 1 if match.group('tz_offset_sign') == '+' else -1
            tzinfo = FixedOffset(tz_sign * (tz_hour * 60 + tz_minute))
        else:
            tzinfo = utc
        microsecond = match.group('microsecond') or '0'
        microsecond += '0' * (6 - len(microsecond))
        try:
            dt = datetime.datetime(
                int(match.group('year')),
                int(match.group('month')),
                int(match.group('day')),
                int(match.group('hour')),
                int(match.group('minute')),
                int(match.group('second')),
                int(microsecond),
                tzinfo=tzinfo
            )
        except ValueError as e:
            raise DecodeError(e)
        if self.prefer_utc and tzinfo is not utc:
            dt = dt.astimezone(utc)
        return dt


class Rfc822(Codec):
    """Codec to encode/decode :class:`datetime.datetime` values to/from
    :rfc:`822` format.

    :param microseconds: whether to preserve and parse microseconds as well.
                         :const:`False` by default since it's not standard
                         compliant
    :type microseconds: :class:`bool`

    .. versionadded:: 0.3.0
       Added ``microseconds`` option.

    """

    WEEKDAYS = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
    MONTHS = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug',
              'Sep', 'Oct', 'Nov', 'Dec')
    TIMEZONES = {
        'UT': FixedOffset(0 * 60),
        'UTC': FixedOffset(0 * 60),
        'GMT': FixedOffset(0 * 60),
        "EST": FixedOffset(-5 * 60),
        "EDT": FixedOffset(-4 * 60),
        "CST": FixedOffset(-6 * 60),
        "CDT": FixedOffset(-5 * 60),
        "MST": FixedOffset(-7 * 60),
        "MDT": FixedOffset(-6 * 60),
        "PST": FixedOffset(-8 * 60),

        "1A": FixedOffset(-1 * 60),
        "1B": FixedOffset(-2 * 60),
        "1C": FixedOffset(-3 * 60),
        "1D": FixedOffset(-4 * 60),
        "1E": FixedOffset(-5 * 60),
        "1F": FixedOffset(-6 * 60),
        "1G": FixedOffset(-7 * 60),
        "1H": FixedOffset(-8 * 60),
        "1I": FixedOffset(-9 * 60),
        "1K": FixedOffset(-10 * 60),
        "1L": FixedOffset(-11 * 60),
        "1M": FixedOffset(-12 * 60),
        "1N": FixedOffset(1 * 60),
        "1O": FixedOffset(2 * 60),
        "1P": FixedOffset(3 * 60),
        "1Q": FixedOffset(4 * 60),
        "1R": FixedOffset(5 * 60),
        "1S": FixedOffset(6 * 60),
        "1T": FixedOffset(7 * 60),
        "1U": FixedOffset(8 * 60),
        "1V": FixedOffset(9 * 60),
        "1W": FixedOffset(10 * 60),
        "1X": FixedOffset(11 * 60),
        "1Y": FixedOffset(12 * 60),
        "1Z": FixedOffset(0 * 60),
    }
    PATTERN = re.compile(r'''
        ^ \s*
        (?:Sun|Mon|Tue|Wed|Thu|Fri|Sat ) , \s+
        (?P<day> \d\d? ) \s+
        (?P<month> ''' + '|'.join(MONTHS) + r''' ) \s+
        (?P<year> \d{4} ) \s+
        (?P<hour> \d\d ) : (?P<minute> \d\d ) : (?P<second> \d\d )
        (?: \. (?P<ms> \d{6} ) )? \s+
        (?P<tz> (?P<tz_offset> (?P<tz_offset_sign> [\+\-] )
                               (?P<tz_offset_hour> [0-9]{2} ) :?
                               (?P<tz_offset_minute> [0-9]{2} ) )
        |       (?P<tz_named>''' + '|'.join(map(re.escape, TIMEZONES)) + r''' )
        )
        \s* $
    ''', re.IGNORECASE | re.VERBOSE)

    def __init__(self, microseconds=False):
        super(Rfc822, self).__init__()
        self.microseconds = bool(microseconds)
        if microseconds:
            if IRON_PYTHON:
                # IronPython strftime() seems to ignore %f
                fmt = ('{w}, {t:%d} {m} {t:%Y %H:%M:%S}.{t.microsecond:06} '
                       '{tz_h:+03d}{tz_m:02d}')
            else:
                fmt = ('{w}, {t:%d} {m} {t:%Y %H:%M:%S}.{t:%f} '
                       '{tz_h:+03d}{tz_m:02d}')
        else:
            fmt = '{w}, {t:%d} {m} {t:%Y %H:%M:%S} {tz_h:+03d}{tz_m:02d}'
        self._format = fmt.format

    def encode(self, value):
        if not isinstance(value, datetime.datetime):
            raise EncodeError(
                '{0.__module__}.{0.__name__} accepts only datetime.datetime '
                'value, not {1!r}'.format(type(self), value)
            )
        elif value.tzinfo is None:
            raise EncodeError(
                '{0.__module__}.{0.__name__} does not accept naive datetime.'
                'datetime value, but {1!r} lacks tzinfo attribute'.format(
                    type(self), value
                )
            )
        offset = value.tzinfo.utcoffset(value)
        minutes = offset.seconds // 60
        res = self._format(
            t=value,
            w=self.WEEKDAYS[value.weekday()],
            m=self.MONTHS[value.month - 1],
            tz_h=minutes // 60,
            tz_m=minutes % 60
        )
        return res

    def decode(self, text):
        exc = DecodeError(repr(text) + ' is an invalid rfc822 string')
        m = self.PATTERN.match(text)
        if not m:
            raise exc
        day = int(m.group('day'))
        month_string = m.group('month')
        try:
            month = self.MONTHS.index(month_string)
        except ValueError:
            raise exc
        else:
            month += 1
        year = int(m.group('year'))
        hour = int(m.group('hour'))
        minute = int(m.group('minute'))
        second = int(m.group('second'))
        if self.microseconds:
            ms = int((m.group('ms') or '').lstrip('0') or '0')
        else:
            ms = 0
        if m.group('tz_offset'):
            tz = FixedOffset(
                int(m.group('tz_offset_hour')) * 60 +
                int(m.group('tz_offset_minute')) *
                (1 if m.group('tz_offset_sign') == '+' else -1)
            )
        elif m.group('tz_named'):
            tz = self.TIMEZONES[m.group('tz_named')]
        return datetime.datetime(
            year, month, day,
            hour, minute, second, ms,
            tzinfo=tz
        )


class Integer(Codec):
    """Codec to encode and decode integer numbers."""

    def encode(self, value):
        if not isinstance(value, numbers.Integral):
            raise EncodeError('expected integer, not ' + repr(value))
        return str(int(value))

    def decode(self, text):
        try:
            return int(text)
        except ValueError as e:
            raise DecodeError(str(e))


class Boolean(Codec):
    """Codec to interpret boolean representation in strings e.g. ``'true'``,
    ``'no'``, and encode :class:`bool` values back to string.

    :param true: text to parse as :const:`True`.  ``'true'`` by default
    :type true: :class:`str`, :class:`tuple`
    :param false: text to parse as :const:`False`.  ``'false'`` by default
    :type false: :class:`str`, :class:`tuple`
    :param default_value: default value when it cannot be parsed
    :type default_value: :class:`bool`

    """

    def __init__(self, true="true", false="false", default_value=None):
        self.true = true
        self.false = false
        self.default_value = default_value

    def encode(self, value):
        if value is None:
            value = self.default_value

        if not isinstance(value, bool) and value is not None:
            raise EncodeError("type of {0} must be bool".format(value))

        true = (self.true if isinstance(self.true, string_type)
                else self.true[0])
        false = (self.false if isinstance(self.true, string_type)
                 else self.false[0])

        if value is True:
            return true
        elif value is False:
            return false
        else:
            return None

    def decode(self, text):
        true = (self.true if not isinstance(self.true, string_type)
                else [self.true])
        false = (self.false if not isinstance(self.false, string_type)
                 else [self.false])

        if text in true:
            value = True
        elif text in false:
            value = False
        elif not text:
            value = self.default_value
        else:
            raise DecodeError('invalid string')
        return value


# Workaround for http://bugs.python.org/issue7980
if platform.python_implementation() == 'CPython':
    __import__('_strptime')



""":mod:`libearth.feed` --- Feeds
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:mod:`libearth` internally stores archive data as Atom format.  It's exactly
not a complete set of :rfc:`4287`, but a subset of the most of that.
Since it's not intended for crawling but internal representation, it does not
follow robustness principle or such thing.  It simply treats stored data are
all valid and well-formed.

"""
import cgi
import collections
import re

from .codecs import Boolean, Enum, Rfc3339
from .compat import UNICODE_BY_DEFAULT, string_type, text_type
from .sanitizer import clean_html, sanitize_html
from .session import MergeableDocumentElement
from .schema import (Attribute, Child, Content as ContentValue, DocumentElement,
                     Element, Text as TextChild, element_list_for)
from .tz import now

__all__ = ('ATOM_XMLNS', 'MARK_XMLNS', 'Category', 'Content', 'Entry', 'Feed',
           'Generator', 'Link', 'LinkList', 'Mark', 'Metadata', 'Person',
           'Source', 'Text')


#: (:class:`str`) The XML namespace name used for Atom (:rfc:`4287`).
ATOM_XMLNS = 'http://www.w3.org/2005/Atom'

#: (:class:`str`) The XML namespace name used for Earth Reader :class:`Mark`
#: metadata.
MARK_XMLNS = 'http://earthreader.org/mark/'


class Text(Element):
    """Text construct defined in :rfc:`4287#section-3.1` (section 3.1)."""

    #: (:class:`str`) The type of the text.  It could be one of ``'text'``
    #: or ``'html'``.  It corresponds to :rfc:`4287#section-3.1.1` (section
    #: 3.1.1).
    #:
    #: .. note::
    #:
    #:    It currently does not support ``'xhtml'``.
    type = Attribute('type', Enum(['text', 'html']),
                     required=True, default=lambda _: 'text')

    #: (:class:`str`) The content of the text.  Interpretation for this
    #: has to differ according to its :attr:`type`.  It corresponds to
    #: :rfc:`4287#section-3.1.1.1` (section 3.1.1.1) if :attr:`type` is
    #: ``'text'``, and :rfc:`4287#section-3.1.1.2` (section 3.1.1.2) if
    #: :attr:`type` is ``'html'``.
    value = ContentValue()

    @property
    def sanitized_html(self):
        """(:class:`str`) The secure HTML string of the text.  If it's
        a plain text, this becomes entity-escaped HTML string (for example,
        ``'<Hello>'`` becomes ``'&lt;Hello&gt;'``), and if it's a HTML text,
        the ``value`` is sanitized (for example,
        ``'<script>alert(1);</script><p>Hello</p>'`` comes ``'<p>Hello</p>'``).

        """
        if self.type == 'html':
            return sanitize_html(self.value)
        elif self.type == 'text':
            return cgi.escape(self.value, quote=True).replace('\n', '<br>\n')

    @classmethod
    def __coerce_from__(cls, value):
        if isinstance(value, string_type):
            return cls(value=value, type='text')
        raise TypeError('expected a string or an instance of {0.__module__}.'
                        '{0.__name__}, not {1!r}'.format(cls, value))

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.type == other.type and self.value == other.value)

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        if not self.value:
            return ''
        elif self.type == 'html':
            return clean_html(self.value)
        elif self.type == 'text':
            return self.value

    if UNICODE_BY_DEFAULT:
        __str__ = __unicode__
    else:
        __str__ = lambda self: unicode(self).encode('utf-8')

    def __html__(self):
        return self.sanitized_html

    def __repr__(self):
        return '{0.__module__}.{0.__name__}(type={1!r}, value={2!r})'.format(
            type(self), self.type, self.value
        )


class Person(Element):
    """Person construct defined in :rfc:`4287#section-3.2` (section 3.2)."""

    #: (:class:`str`) The human-readable name for the person.  It corresponds
    #: to ``atom:name`` element of :rfc:`4287#section-3.2.1` (section 3.2.1).
    name = TextChild('name', xmlns=ATOM_XMLNS, required=True)

    #: (:class:`str`) The optional URI associated with the person.
    #: It corresponds to ``atom:uri`` element of :rfc:`4287#section-3.2.2`
    #: (section 3.2.2).
    uri = TextChild('uri', xmlns=ATOM_XMLNS)

    #: (:class:`str`) The optional email address associated with the person.
    #: It corresponds to ``atom:email`` element of :rfc:`4287#section-3.2.3`
    #: (section 3.2.3).
    email = TextChild('email', xmlns=ATOM_XMLNS)

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.name == other.name and
                self.uri == other.uri and
                self.email == other.email)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self.uri, self.email))

    def __unicode__(self):
        ref = self.uri or self.email
        if ref:
            return text_type('{0} <{1}>').format(self.name, ref)
        return self.name

    if UNICODE_BY_DEFAULT:
        __str__ = __unicode__
    else:
        __str__ = lambda self: unicode(self).encode('utf-8')

    def __html__(self):
        name = cgi.escape(self.name, quote=True)
        ref = self.uri or self.email and 'mailto:' + self.email
        if ref:
            return text_type('<a href="{1}">{0}</a>').format(
                name,
                cgi.escape(ref, quote=True)
            )
        return name

    def __repr__(self):
        return ('{0.__module__}.{0.__name__}(name={1!r}, uri={2!r}'
                ', email={3!r})').format(type(self), self.name, self.uri,
                                         self.email)


class Link(Element):
    """Link element defined in :rfc:`4287#section-4.2.7` (section 4.2.7)."""

    #: (:class:`str`) The link's required URI.  It corresponds to ``href``
    #: attribute of :rfc:`4287#section-4.2.7.1` (section 4.2.7.1).
    uri = Attribute('href', required=True)

    #: (:class:`str`) The relation type of the link.  It corresponds to
    #: ``rel`` attribute of :rfc:`4287#section-4.2.7.2` (section 4.2.7.2).
    #:
    #: .. seealso::
    #:
    #:    `Existing rel values`__ --- Microformats Wiki
    #:        This page contains tables of known HTML ``rel`` values from
    #:        specifications, formats, proposals, brainstorms, and non-trivial
    #:        POSH_ usage in the wild.  In addition, dropped and rejected
    #:        values are listed at the end for comprehensiveness.
    #:
    #:    __ http://microformats.org/wiki/existing-rel-values
    #:    .. _POSH: http://microformats.org/wiki/POSH
    relation = Attribute('rel', required=True, default=lambda _: 'alternate')

    #: (:class:`str`) The optional hint for the MIME media type of the linked
    #: content.  It corresponds to ``type`` attribute of
    #: :rfc:`4287#section-4.2.7.3` (section 4.2.7.3).
    mimetype = Attribute('type')

    #: (:class:`str`) The language of the linked content.  It corresponds
    #: to ``hreflang`` attribute of :rfc:`4287#section-4.2.7.4` (section
    #: 4.2.7.4).
    language = Attribute('hreflang')

    #: (:class:`str`) The title of the linked resource.  It corresponds to
    #: ``title`` attribute of :rfc:`4287#section-4.2.7.5` (section 4.2.7.5).
    title = Attribute('title')

    #: (:class:`numbers.Integral`) The optional hint for the length of
    #: the linked content in octets.  It corresponds to ``length`` attribute
    #: of :rfc:`4287#section-4.2.7.6` (section 4.2.7.6).
    byte_size = Attribute('length')

    def __eq__(self, other):
        return (isinstance(other, type(self)) and
                self.uri == other.uri and
                self.relation == other.relation and
                self.mimetype == other.mimetype and
                self.language == other.language and
                self.title == other.title and
                self.byte_size == other.byte_size)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.uri, self.relation, self.mimetype, self.language,
                     self.title, self.byte_size))

    @property
    def html(self):
        """(:class:`bool`) Whether its :attr:`mimetype` is HTML (or XHTML).

        .. versionadded:: 0.2.0

        """
        if self.mimetype:
            match = re.match(r'^\s*([^;/\s]+/[^;/\s]+)\s*(?:;\s*.*)?$',
                             self.mimetype)
            if match:
                mimetype = match.group(1)
                return mimetype in ('text/html', 'application/xhtml+xml')
        return False

    def __unicode__(self):
        return self.uri

    if UNICODE_BY_DEFAULT:
        __str__ = __unicode__
    else:
        __str__ = lambda self: unicode(self).encode('utf-8')

    def __html__(self):
        mapping = [
            ('rel', self.relation),
            ('type', self.mimetype),
            ('hreflang', self.language),
            ('href', self.uri),
            ('title', self.title)
        ]
        return text_type('<link{0}>').format(
            text_type('').join(
                text_type(' {0}="{1}"').format(attr, value)
                for attr, value in mapping if value
            )
        )

    def __repr__(self):
        return ('{0.__module__}.{0.__name__}(uri={1!r}, relation={2!r}'
                ', mimetype={3!r}, language={4!r}, title={5!r}'
                ', byte_size={6!r})').format(type(self), self.uri,
                                             self.relation, self.mimetype,
                                             self.language, self.title,
                                             self.byte_size)


@element_list_for(Link)
class LinkList(collections.MutableSequence):
    """Element list mixin specialized for :class:`Link`."""

    def filter_by_mimetype(self, pattern):
        """Filter links by their :attr:`~Link.mimetype` e.g.::

            links.filter_by_mimetype('text/html')

        ``pattern`` can include wildcards (``*``) as well e.g.::

            links.filter_by_mimetype('application/xml+*')

        :param pattern: the mimetype pattern to filter
        :type pattern: :class:`str`
        :returns: the filtered links
        :rtype: :class:`LinkList`

        """
        if '*' in pattern:
            regex = re.compile(
                '.+?'.join(re.escape(s) for s in pattern.split('*')) + '$'
            )
            return LinkList.list_type(
                link for link in self
                if link.mimetype and regex.match(link.mimetype)
            )
        return LinkList.list_type(l for l in self if l.mimetype == pattern)

    @property
    def permalink(self):
        """(:class:`Link`) Find the permalink from the list.  The following
        list shows precedence of lookup conditions:

        1. :attr:`~Link.html`, and :attr:`~Link.relation` is ``'alternate'``
        2. :attr:`~Link.html`
        3. :attr:`~Link.relation` is ``'alternate'``
        4. No permalink: return :const:`None`

        .. versionadded:: 0.2.0

        """
        links = [(lnk, (lnk.html, lnk.relation == 'alternate')) for lnk in self]
        filterred_links = [(l, cond) for l, cond in links if cond[0] or cond[1]]
        try:
            link, _ = max(filterred_links, key=lambda pair: pair[1])
        except ValueError:
            pass
        else:
            return link

    @property
    def favicon(self):
        """(:class:`Link`) Find the link to a favicon, also known as
        a shortcut or bookmark icon, if it exists.

        .. versionadded:: 0.3.0

        """
        for link in self:
            if 'icon' in link.relation.split():
                return link


# FIXME: it probably would be common for all specialized element list types
LinkList.list_type = type('LinkList.list_type', (list, LinkList), {})


class Category(Element):
    """Category element defined in :rfc:`4287#section-4.2.2` (section 4.2.2)."""

    #: (:class:`str`) The required machine-readable identifier string of
    #: the cateogry.  It corresponds to ``term`` attribute of
    #: :rfc:`4287#section-4.2.2.1` (section 4.2.2.1).
    term = Attribute('term', required=True)

    #: (:class:`str`) The URI that identifies a categorization scheme.
    #: It corresponds to ``scheme`` attribute of :rfc:`4287#section-4.2.2.2`
    #: (section 4.2.2.2).
    #:
    #: .. seealso::
    #:
    #:    - `Tag Scheme?`__ by Tim Bray
    #:    - `Representing tags in Atom`__ by Edward O'Connor
    #:
    #:    __ http://www.tbray.org/ongoing/When/200x/2007/02/01/Tag-Scheme
    #:    __ http://edward.oconnor.cx/2007/02/representing-tags-in-atom
    scheme_uri = Attribute('scheme')

    #: (:class:`str`) The optional human-readable label for display in
    #: end-user applications.  It corresponds to ``label`` attribute of
    #: :rfc:`4287#section-4.2.2.3` (section 4.2.2.3).
    label = Attribute('label')

    def __entity_id__(self):
        return self.term

    def __merge_entities__(self, other):
        if not self.label:
            self.label = other.label
        return self

    def __unicode__(self):
        return self.label or self.term

    if UNICODE_BY_DEFAULT:
        __str__ = __unicode__
    else:
        __str__ = lambda self: unicode(self).encode('utf-8')

    def __repr__(self):
        return ('{0.__module__}.{0.__name__}(term={1!r}, scheme_uri={2!r}'
                ', label={3!r})').format(type(self), self.term,
                                         self.scheme_uri, self.label)


class Content(Text):
    """Content construct defined in :rfc:`4287#section-4.1.3`
    (section 4.1.3).

    """

    #: (:class:`collections.Mapping`) The mapping of :attr:`type` string
    #: (e.g. ``'text'``) to the corresponding MIME type
    #: (e.g. :mimetype:`text/plain`).
    TYPE_MIMETYPE_MAP = {
        'text': 'text/plain',
        'html': 'text/html',
        'xhtml': 'application/xhtml+xml'
    }

    #: (:class:`re.RegexObject`) The regular expression pattern that matches
    #: with valid MIME type strings.
    MIMETYPE_PATTERN = re.compile(r'''
        ^
        (?P<type> [A-Za-z0-9!#$&.+^_-]{1,127} )
        /
        (?P<subtype> [A-Za-z0-9!#$&.+^_-]{1,127} )
        $
    ''', re.VERBOSE)

    #: (:class:`str`) An optional remote content URI to retrieve the content.
    source_uri = Attribute('src')

    @property
    def mimetype(self):
        """(:class:`str`) The mimetype of the content."""
        try:
            mimetype = self.TYPE_MIMETYPE_MAP[self.type]
        except KeyError:
            mimetype = self.type
        if self.MIMETYPE_PATTERN.match(mimetype):
            return mimetype
        raise ValueError(repr(mimetype) + ' is invalid mimetype')

    @mimetype.setter
    def mimetype(self, mimetype):
        match = self.MIMETYPE_PATTERN.match(mimetype)
        if not match:
            raise ValueError(repr(mimetype) + ' is invalid mimetype')
        if match.group('type') == 'text':
            subtype = match.group('subtype')
            if subtype == 'plain':
                self.type = 'text'
                return
            elif subtype == 'html':
                self.type = 'html'
                return
        self.type = mimetype

    def __eq__(self, other):
        if self.source_uri:
            return (self.mimetype == other.mimetype and
                    self.source_uri == other.source_uri)
        return self.mimetype == other.mimetype and self.value == other.value

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        if not self.source_uri:
            return super(Content, self).__repr__()
        format_string = ('{0.__module__}.{0.__name__}'
                         '(mimetype={1!r}, source_uri={2!r})')
        return format_string.format(type(self), self.mimetype, self.source_uri)


class Generator(Element):
    """Identify the agent used to generate a feed, for debugging and
    other purposes.  It's corresponds to ``atom:generator`` element
    of :rfc:`4287#section-4.2.4` (section 4.2.4).

    """

    #: (:class:`str`) A URI that represents something relavent to the agent.
    uri = Attribute('uri')

    #: (:class:`str`) The version of the generating agent.
    version = Attribute('version')

    #: (:class:`str`) The human-readable name for the generating agent.
    value = ContentValue()

    def __eq__(self, other):
        return (self.uri == other.uri and self.version == other.version and
                self.value == other.value)

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        if self.version:
            return '{0} {1}'.format(self.value, self.version)
        return self.value

    if UNICODE_BY_DEFAULT:
        __str__ = __unicode__
    else:
        __str__ = lambda self: unicode(self).encode('utf-8')

    def __html__(self):
        label = cgi.escape(self.value)
        if self.version:
            label = '{0} {1}'.format(label, cgi.escape(self.version))
        if self.uri:
            return '<a href="{0}">{1}</a>'.format(cgi.escape(self.uri), label)
        return label

    def __repr__(self):
        return '{0.__module__}.{0.__name__}(value={1!r}, version={2!r}, uri=' \
               '{3!r})'.format(type(self), self.value, self.version, self.uri)


class Metadata(Element):
    """Common metadata shared by :class:`Source`, :class:`Entry`, and
    :class:`Feed`.

    """

    #: (:class:`str`) The URI that conveys a permanent, universally unique
    #: identifier for an entry or feed.  It corresponds to ``atom:id``
    #: element of :rfc:`4287#section-4.2.6` (section 4.2.6).
    id = TextChild('id', xmlns=ATOM_XMLNS, required=True)

    #: (:class:`Text`) The human-readable title for an entry or feed.
    #: It corresponds to ``atom:title`` element of :rfc:`4287#section-4.2.14`
    #: (section 4.2.14).
    title = Child('title', Text, xmlns=ATOM_XMLNS, required=True)

    #: (:class:`collections.LinkList`) The list of :class:`Link` objects
    #: that define a reference from an entry or feed to a web resource.
    #: It corresponds to ``atom:link`` element of :rfc:`4287#section-4.2.7`
    #: (section 4.2.7).
    links = Child('link', Link, xmlns=ATOM_XMLNS, multiple=True)

    #: (:class:`datetime.datetime`) The tz-aware :class:`~datetime.datetime`
    #: indicating the most recent instant in time when the entry was modified
    #: in a way the publisher considers significant.  Therefore, not all
    #: modifications necessarily result in a changed :attr:`updated_at` value.
    #: It corresponds to ``atom:updated`` element of :rfc:`4287#section-4.2.15`
    #: (section 4.2.15).
    updated_at = TextChild('updated', Rfc3339, xmlns=ATOM_XMLNS, required=True)

    #: (:class:`collections.MutableSequence`) The list of :class:`Person`
    #: objects which indicates the author of the entry or feed.  It corresponds
    #: to ``atom:author`` element of :rfc:`4287#section-4.2.1` (section 4.2.1).
    authors = Child('author', Person, xmlns=ATOM_XMLNS, multiple=True)

    #: (:class:`collections.MutableSequence`) The list of :class:`Person`
    #: objects which indicates a person or other entity who contributed to
    #: the entry or feed.  It corresponds to ``atom:contributor`` element  of
    #: :rfc:`4287#section-4.2.3` (section 4.2.3).
    contributors = Child('contributor', Person, xmlns=ATOM_XMLNS, multiple=True)

    #: (:class:`collections.MutableSequence`) The list of :class:`Category`
    #: objects that conveys information about categories associated with
    #: an entry or feed.  It corresponds to ``atom:category`` element of
    #: :rfc:`4287#section-4.2.2` (section 4.2.2).
    categories = Child('category', Category, xmlns=ATOM_XMLNS, multiple=True)

    #: (:class:`Text`) The text field that conveys information about rights
    #: held in and of an entry or feed.  It corresponds to ``atom:rights``
    #: element of :rfc:`4287#section-4.2.10` (section 4.2.10).
    rights = Child('rights', Text, xmlns=ATOM_XMLNS)

    def __entity_id__(self):
        return self.id

    def __unicode__(self):
        return unicode(self.title) if self.title else unicode()

    def __str__(self):
        return str(self.title) if self.title else ''

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} {1} {2!r}>'.format(
            type(self),
            self.id,
            str(self.title)
        )


class Source(Metadata):
    """All metadata for :class:`Feed` excepting :attr:`Feed.entries`.
    It corresponds to ``atom:source`` element of :rfc:`4287#section-4.2.10`
    (section 4.2.10).

    """

    #: (:class:`Text`) A text that conveys a human-readable description or
    #: subtitle for a feed.  It corresponds to ``atom:subtitle`` element of
    #: :rfc:`4287#section-4.2.12` (section 4.2.12).
    subtitle = Child('subtitle', Text, xmlns=ATOM_XMLNS)

    #: (:class:`Generator`) Identify the agent used to generate a feed,
    #: for debugging and other purposes.  It corresponds to
    #: ``atom:generator`` element of :rfc:`4287#section-4.2.4` (section 4.2.4).
    generator = Child('generator', Generator, xmlns=ATOM_XMLNS)

    #: (:class:`str`) URI that identifies an image that provides visual
    #: identification for a feed.  It corresponds to ``atom:logo`` element
    #: of :rfc:`4287#section-4.2.8` (section 4.2.8).
    logo = TextChild('logo', xmlns=ATOM_XMLNS)

    #: (:class:`str`) URI that identifies an image that provides iconic
    #: visual identification for a feed.  It corresponds to ``atom:icon``
    #: element of :rfc:`4287#section-4.2.5` (section 4.2.5).
    icon = TextChild('icon', xmlns=ATOM_XMLNS)


class Mark(Element):
    """Represent whether the entry is read, starred, or tagged by user.
    It's not a part of :rfc:`4287` Atom standard, but extension for
    Earth Reader.

    """

    #: (:class:`bool`) Whether it's marked or not.
    marked = ContentValue(Boolean)

    #: (:class:`datetime.datetime`) Updated time.
    updated_at = Attribute('updated', Rfc3339)

    def __bool__(self):
        return bool(self.marked)

    __nonzero__ = __bool__

    def __eq__(self, other):
        return (isinstance(other, Mark) and
                bool(self) == bool(other) and
                self.updated_at == other.updated_at)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return bool((bool(self), self.updated_at))

    def __entity_id__(self):
        """If there are two or more marks that have the same tag name, these
        are all should be merged into one.

        """
        return 1

    def __merge_entities__(self, other):
        return max(self, other, key=lambda mark: mark.updated_at)

    @classmethod
    def __coerce_from__(cls, value):
        if isinstance(value, bool):
            return cls(marked=value, updated_at=now())
        raise TypeError('expected bool or an instance of {0.__module__}.'
                        '{0.__name__}, not {1!r}'.format(cls, value))

    def __repr__(self):
        fmt = '{0.__module__}.{0.__name__}(marked={1!r}, updated_at={2!r})'
        return fmt.format(type(self), self.marked, self.updated_at)


class Entry(DocumentElement, Metadata):
    """Represent an individual entry, acting as a container for metadata and
    data associated with the entry.  It corresponds to ``atom:entry`` element
    of :rfc:`4287#section-4.1.2` (section 4.1.2).

    """

    __tag__ = 'entry'
    __xmlns__ = ATOM_XMLNS

    #: (:class:`datetime.datetime`) The tz-aware :class:`~datetime.datetime`
    #: indicating an instant in time associated with an event early in the
    #: life cycle of the entry.  Typically, :attr:`published_at` will be
    #: associated with the initial creation or first availability of
    #: the resource.  It corresponds to ``atom:published`` element of
    #: :rfc:`4287#section-4.2.9` (section 4.2.9).
    published_at = TextChild('published', Rfc3339, xmlns=ATOM_XMLNS)

    #: (:class:`Text`) The text field that conveys a short summary, abstract,
    #: or excerpt of the entry.  It corresponds to ``atom:summary`` element
    #: of :rfc:`4287#section-4.2.13` (section 4.2.13).
    summary = Child('summary', Text, xmlns=ATOM_XMLNS)

    #: (:class:`Content`) It either contains or links to the content of
    #: the entry.
    #:
    #: It corresponds to ``atom:content`` element of :rfc:`4287#section-4.1.3`
    #: (section 4.1.3).
    content = Child('content', Content, xmlns=ATOM_XMLNS)

    #: (:class:`Source`) If an entry is copied from one feed into another
    #: feed, then the source feed's metadata may be preserved within
    #: the copied entry by adding :attr:`source` if it is not already present
    #: in the entry, and including some or all of the source feed's metadata
    #: as the :attr:`source`'s data.
    #:
    #: It is designed to allow the aggregation of entries from different feeds
    #: while retaining information about an entry's source feed.
    #:
    #: It corresponds to ``atom:source`` element of :rfc:`4287#section-4.2.10`
    #: (section 4.2.10).
    source = Child('source', Source, xmlns=ATOM_XMLNS)

    #: (:class:`Mark`) Whether and when it's read or unread.
    read = Child('read', Mark, xmlns=MARK_XMLNS)

    #: (:class:`Mark`) Whether and when it's starred or unstarred.
    starred = Child('starred', Mark, xmlns=MARK_XMLNS)

    def __merge_entities__(self, other):
        for attribute in 'read', 'starred':
            self_mark = getattr(self, attribute)
            other_mark = getattr(other, attribute)
            if self_mark is None:
                setattr(self, attribute, other_mark)
            elif self_mark is not None and other_mark is not None:
                merged_mark = self_mark.__merge_entities__(other_mark)
                setattr(self, attribute, merged_mark)
        return self


class Feed(MergeableDocumentElement, Source):
    """Atom feed document, acting as a container for metadata and data
    associated with the feed.

    It corresponds to ``atom:feed`` element of :rfc:`4287#section-4.1.1`
    (section 4.1.1).

    """

    __tag__ = 'feed'
    __xmlns__ = ATOM_XMLNS

    #: (:class:`collections.MutableSequence`) The list of :class:`Entry` objects
    #: that represent an individual entry, acting as a container for metadata
    #: and data associated with the entry.
    #: It corresponds to ``atom:entry`` element of :rfc:`4287#section-4.1.2`
    #: (section 4.1.2).
    entries = Child(
        'entry', Entry,
        xmlns=ATOM_XMLNS,
        multiple=True,
        sort_key=lambda e: e.published_at or e.updated_at,
        sort_reverse=True
    )



""":mod:`libearth.compat` --- Compatibility layer
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides several subtle things to support
multiple Python versions (2.6, 2.7, 3.2, 3.3) and VM implementations
(CPython, PyPy).

"""
import functools
import io
import platform
import sys
import types

__all__ = ('IRON_PYTHON', 'PY3', 'UNICODE_BY_DEFAULT', 'binary', 'binary_type',
           'encode_filename', 'file_types', 'string_type', 'text', 'text_type',
           'xrange')


#: (:class:`bool`) Whether it is Python 3.x or not.
PY3 = sys.version_info >= (3,)

#: (:class:`bool`) Whether it is IronPython or not.
IRON_PYTHON = platform.python_implementation() == 'IronPython'

#: (:class:`bool`) Whether the Python VM uses Unicode strings by default.
#: It must be :const:`True` if :const:`PY3` or IronPython.
UNICODE_BY_DEFAULT = PY3 or IRON_PYTHON

#: (:class:`type`) Type for representing binary data.  :class:`str` in Python 2
#: and :class:`bytes` in Python 3.
binary_type = bytes if PY3 or IRON_PYTHON else str

#: (:class:`type`) Type for text data.  :class:`basestring` in Python 2
#: and :class:`str` in Python 3.
string_type = str if PY3 else basestring

#: (:class:`type`) Type for representing Unicode textual data.
#: :class:`unicode` in Python 2 and :class:`str` in Python 3.
text_type = str if PY3 else unicode

#: The function that takes a Unicode codepoint and returns the corresponding
#: Unicode character.
unichr = chr if PY3 else unichr


def binary(string, var=None):
    """Makes ``string`` to :class:`str` in Python 2.
    Makes ``string`` to :class:`bytes` in Python 3 or IronPython.

    :param string: a string to cast it to :data:`binary_type`
    :type string: :class:`bytes`, :class:`str`, :class:`unicode`
    :param var: an optional variable name to be used for error message
    :type var: :class:`str`

    """
    if isinstance(string, text_type):
        if IRON_PYTHON:
            return bytes(string, sys.getdefaultencoding())
        return string.encode()
    elif isinstance(string, binary_type):
        return string
    if var:
        raise TypeError('{0} must be a string, not {1!r}'.format(var, string))
    raise TypeError('expected a string, not ' + repr(string))


def text(string):
    """Makes ``string`` to :class:`str` in Python 3 or IronPython.
    Does nothing in Python 2.

    :param string: a string to cast it to :data:`text_type`
    :type string: :class:`bytes`, :class:`str`, :class:`unicode`

    """
    return string


if PY3 or IRON_PYTHON:
    def text(string):
        if isinstance(string, bytes):
            return string.decode('utf-8')
        return string


#: The :func:`xrange()` function.  Alias for :func:`range()` in Python 3.
xrange = range if PY3 else xrange


#: The :func:`reduce()` function.  Alias for :func:`functools.reduce()`
#: in Python 3.
reduce = functools.reduce if PY3 else reduce


#: (:class:`type`, :class:`tuple`) Types for file objects that have
#: ``fileno()``.
file_types = io.RawIOBase if PY3 else (io.RawIOBase, types.FileType)


def encode_filename(filename):
    """If ``filename`` is a :data:`text_type`, encode it to
    :data:`binary_type` according to filesystem's default encoding.

    """
    if isinstance(filename, text_type):
        return filename.encode(sys.getfilesystemencoding())
    return filename



""":mod:`libearth.version` --- Version data
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from __future__ import print_function

#: (:class:`tuple`) The triple of version numbers e.g. ``(1, 2, 3)``.
VERSION_INFO = (0, 3, 1)

#: (:class:`str`) The version string e.g. ``'1.2.3'``.
VERSION = '{0}.{1}.{2}'.format(*VERSION_INFO)


if __name__ == '__main__':
    print(VERSION)



""":mod:`libearth.stage` --- Staging updates and transactions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Stage is a similar concept to Git's one.  It's a unit of updates,
so every change to the repository should be done through a stage.

It also does more than Git's stage: :class:`Route`.  Routing system
hides how document should be stored in the repository, and provides
the natural object-mapping interface instead.

Stage also provides transactions.  All operations on staged documents should
be done within a transaction.  You can open and close a transaction using
:keyword:`with` statement e.g.::

    with stage:
        subs = stage.subscriptions
        stage.subscriptions = some_operation(subs)

Transaction will merge all simultaneous updates if there are multiple updates
when it's committed.  You can easily achieve thread safety using transactions.

Note that it however doesn't guarantee data integrity between multiple
processes, so *you have to use different session ids when there are multiple
processes.*

"""
import collections
import contextlib
import io
import re
import sys
import threading
import traceback

if sys.version_info >= (3,):
    try:
        import _thread
    except ImportError:
        import _dummy_thread as _thread
else:
    try:
        import thread as _thread
    except ImportError:
        import dummy_thread as _thread

try:
    import greenlet
except ImportError:
    greenlet = None
try:
    import stackless
except ImportError:
    stackless = None

from .compat import IRON_PYTHON, binary_type, reduce
from .feed import Feed
from .repository import Repository, RepositoryKeyError
from .schema import read, write
from .session import (MergeableDocumentElement, RevisionSet, Session,
                      parse_revision)
from .subscribe import SubscriptionList
from .tz import now

__all__ = ('BaseStage', 'Directory', 'DirtyBuffer', 'Route', 'Stage',
           'TransactionError',
           'compile_format_to_pattern', 'get_current_context_id')


def get_current_context_id():
    """Identifies which context it is (greenlet, stackless, or thread).

    :returns: the identifier of the current context

    """
    global get_current_context_id
    if greenlet is not None:
        if stackless is None:
            get_current_context_id = greenlet.getcurrent
            return greenlet.getcurrent()
        return greenlet.getcurrent(), stackless.getcurrent()
    elif stackless is not None:
        get_current_context_id = stackless.getcurrent
        return stackless.getcurrent()
    get_current_context_id = _thread.get_ident
    return _thread.get_ident()


class BaseStage(object):
    """Base stage class that routes nothing yet.  It should be inherited
    to route document types.  See also :class:`Route` class.

    It's a context manager, which is possible to be passed to :keyword:`with`
    statement.  The context maintains a transaction, that is required for
    all operations related to the stage::

        with stage:
            v = stage.some_value
            stage.some_value =  operate(v)

    If any ongoing transaction is not present while the operation requires it,
    it will raise :exc:`TransactionError`.

    :param session: the current session to stage
    :type session: :class:`~libearth.session.Session`
    :param repository: the repository to stage
    :type repository: :class:`~libearth.repository.Repository`

    """

    #: (:class:`collections.Sequence`) The repository key of the directory
    #: where session list are stored.
    SESSION_DIRECTORY_KEY = ['.sessions']

    #: (:class:`~libearth.session.Session`) The current session of the stage.
    session = None

    #: (:class:`~libearth.repository.Repository`) The staged repository.
    repository = None

    #: (:class:`collections.MutableMapping`) Ongoing transactions.  Keys are
    #: the context identifier (that :func:`get_current_context_id()` returns),
    #: and values are pairs of the :class:`DirtyBuffer` that should be written
    #: when the transaction is committed, and stack information.
    transactions = None

    def __init__(self, session, repository):
        if not isinstance(session, Session):
            raise TypeError('session must be an instance of {0.__module__}.'
                            '{0.__name__}, not {1!r}'.format(Session, session))
        elif not isinstance(repository, Repository):
            raise TypeError(
                'repository must be an instance of {0.__module__}.'
                '{0.__name__}, not {1!r}'.format(Repository, repository)
            )
        self.session = session
        self.repository = repository
        self.transactions = {}
        self.lock = threading.RLock()

    def __enter__(self):
        context_id = get_current_context_id()
        transactions = self.transactions
        try:
            pair = transactions[context_id]
        except KeyError:
            pass
        else:
            _, stack = pair
            raise TransactionError(
                'cannot doubly begin transactions for the same context; '
                'please commit the previously begun transaction first.\n'
                'note that previous transaction is begun at:\n' +
                ''.join('  ' + line.replace('\n', '\n  ', 1) for line in stack)
            )
        transactions[context_id] = (DirtyBuffer(self.repository, self.lock),
                                    traceback.format_stack())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        dirty_buffer = self.get_current_transaction(pop=True)
        if exc_type is None:
            dirty_buffer.flush()
        self.touch()

    def get_current_transaction(self, pop=False):
        """Get the current ongoing transaction.  If any transaction is not
        begun yet, it raises :exc:`TransactionError`.

        :returns: the dirty buffer that should be written when the transaction
                  is committed
        :rtype: :class:`DirtyBuffer`
        :raises TransactionError: if not any transaction is not begun yet

        """
        context_id = get_current_context_id()
        trans_dict = self.transactions
        try:
            pair = trans_dict.pop(context_id) if pop else trans_dict[context_id]
        except KeyError:
            raise TransactionError(
                'there is no ongoing transaction for the current context; '
                'please begin the transaction using with keyword e.g.:\n'
                '    stage = {0!r}\n'
                '    with stage:\n'
                '        do_something(stage)\n'.format(self)
            )
        dirty_buffer, _ = pair
        return dirty_buffer

    @property
    def sessions(self):
        """(:class:`collections.Set`) List all sessions associated to
         the :attr:`repository`.  It includes the session of the current stage.

        """
        try:
            identifiers = self.repository.list(self.SESSION_DIRECTORY_KEY)
        except RepositoryKeyError:
            return frozenset()
        return frozenset(Session(identifier=ident) for ident in identifiers)

    def touch(self):
        """Touch the latest staged time of the current :attr:`session`
        into the :attr:`repository`.

        .. note::

           This method is intended to be internal.

        """
        timestamp = now().isoformat()
        if not isinstance(timestamp, binary_type):
            timestamp = binary_type(timestamp, 'ascii')
        self.repository.write(
            self.SESSION_DIRECTORY_KEY + [self.session.identifier],
            [timestamp]
        )

    def read(self, document_type, key):
        """Read a document of ``document_type`` by the given ``key``
        in the staged :attr:`repository`.

        :param document_type:
            the type of document to read. it has to be a subclass of
            :class:`~libearth.session.MergeableDocumentElement`
        :type document_type: :class:`type`
        :param key: the key to find the ``document`` in the :attr:`repository`
        :type key: :class:`collections.Seqeuence`
        :returns: found document instance
        :rtype: :class:`~libearth.session.MergeableDocumentElement`
        :raises libearth.repository.RepositoryKeyError: when the key cannot
                                                        be found

        .. note::

           This method is intended to be internal.  Use routed properties
           rather than this.  See also :class:`Route`.

        """
        if not isinstance(document_type, type):
            raise TypeError('document_type must be a type object, '
                            'not {1!r}'.format(document_type))
        elif not issubclass(document_type, MergeableDocumentElement):
            raise TypeError(
                'document_type must be a subtype of {0.__module__}.'
                '{0.__name__}, not {1.__module__}.{1.__name__}'.format(
                    MergeableDocumentElement,
                    document_type
                )
            )
        repository = self.get_current_transaction()
        chunks = repository.read(key)
        document = read(document_type, chunks)
        assert isinstance(document, MergeableDocumentElement)
        not_stamped = document.__revision__ is None
        if not_stamped:
            return self.write(key, document, merge=False)
        return document

    def read_merged_document(self, document_type, key_spec, key):
        # FIXME: remove assumption that it always takes Session.identifier
        complete_size = len(key)
        pattern = compile_format_to_pattern(key_spec[complete_size])
        for subkey in key_spec[complete_size + 1:]:
            try:
                subkey.format()
            except IndexError:
                raise  # FIXME: should return Directory instead
        repository = self.get_current_transaction()
        docs = []
        for subkey in repository.list(key):
            match = pattern.match(subkey)
            if match:
                k = key + [subkey] + key_spec[complete_size + 1:]
                doc = self.read(document_type, k)
                triple = match.group(1), doc, k
                docs.append(triple)
        session = self.session
        if len(docs) == 1:
            _, doc, __ = docs[0]
            if doc.__revision__.session is session:
                return doc
            doc.__base_revisions__ = doc.__base_revisions__.merge(
                RevisionSet([doc.__revision__])
            )
            doc.__reivison__ = now()
            key = key + [key_spec[complete_size].format(session=session)] \
                      + key_spec[complete_size + 1:]
            return self.write(key, doc, merge=False)
        docs.sort(key=lambda pair: pair[0] == session.identifier,
                  reverse=True)  # the current session comes first
        if docs:
            session_id, doc = reduce(lambda a, b:
                                     (a[0], session.merge(a[1], b[1])), docs)
            return doc

    def write(self, key, document, merge=True):
        """Save the ``document`` to the ``key`` in the staged
        :attr:`repository`.

        :param key: the key to be stored
        :type key: :class:`collections.Sequence`
        :param document: the document to save
        :type document: :class:`~libearth.schema.MergeableDocumentElement`
        :param merge: merge with the previous revision of the same session
                      (if exists).  :const:`True` by default
        :type merge: :class:`bool`
        :returns: actually written document
        :rtype: :class:`~libearth.schema.MergeableDocumentElement`

        .. note::

           This method is intended to be internal.  Use routed properties
           rather than this.  See also :class:`Route`.

        """
        repository = self.get_current_transaction()
        try:
            if not merge:
                raise RepositoryKeyError([])
            prev = repository.read(key)
        except RepositoryKeyError:
            document = self.session.pull(document)
            pull = True
        else:
            prev_doc = read(type(document), prev)
            prev_rev = prev_doc.__revision__
            doc_rev = document.__revision__
            pull = (
                doc_rev is not None and prev_rev is not None and
                (doc_rev.updated_at > prev_rev.updated_at
                 if doc_rev.session is prev_rev.session
                 else document.__base_revisions__.contains(prev_rev))
            )
            if pull:
                # If the document already contains prev_doc, don't merge
                assert prev_rev.session is doc_rev.session
                document = self.session.pull(document)
            else:
                if prev_rev is None:
                    prev_doc = self.session.pull(prev_doc)
                if doc_rev is None:
                    document = self.session.pull(document)
                document = self.session.merge(prev_doc, document, force=True)
        with self.lock:  # FIXME
            bytearray = write(document, canonical_order=True, as_bytes=True)
        repository.write(key, bytearray, _type_hint=type(document))
        return document

    def __repr__(self):
        return '{0.__module__}.{0.__name__}({1!r}, {2!r})'.format(
            type(self), self.session, self.repository
        )


class DirtyBuffer(Repository):
    """Memory-buffered proxy for the repository.  It's used for transaction
    buffer which maintains updates to be written until the ongoing transaction
    is committed.

    :param repository: the bare repository where the buffer will
                       :meth:`flush` to
    :type repository: :class:`~libearth.repository.Repository`
    :param lock: the common lock shared between dirty buffers of the same stage
    :type lock: :class:`threading.RLock`

    .. note::

       This class is intended to be internal.

    """

    #: (:class:`~libearth.repository.Repository`) The bare repository where
    #: the buffer will :meth:`flush` to.
    repository = None

    def __init__(self, repository, lock):
        self.repository = repository
        self.dictionary = {}
        self.lock = lock

    def read(self, key):
        super(DirtyBuffer, self).read(key)
        d = self.dictionary
        for k in key:
            if not isinstance(d, dict):
                raise RepositoryKeyError(key)
            try:
                d = d[k]
            except KeyError:
                with self.lock:
                    return self.repository.read(key)
        return d[1],

    def write(self, key, iterable, _type_hint=None):
        super(DirtyBuffer, self).write(key, iterable)
        d = self.dictionary
        for k in key[:-1]:
            if not isinstance(d, dict):
                raise RepositoryKeyError(key)
            d = d.setdefault(k, {})
        if not isinstance(d, dict):
            raise RepositoryKeyError(key)
        if IRON_PYTHON:
            chunks = io.BytesIO()
            for chunk in iterable:
                chunks.write(chunk)
            bytearray = chunks.getvalue()
        else:
            bytearray = b''.join(iterable)
        d[key[-1]] = _type_hint, bytearray

    def exists(self, key):
        super(DirtyBuffer, self).exists(key)
        d = self.dictionary
        for k in key:
            if not isinstance(d, dict):
                raise RepositoryKeyError(key)
            try:
                d = d[k]
            except KeyError:
                with self.lock:
                    return self.repository.exists(key)
        return True

    def list(self, key):
        super(DirtyBuffer, self).list(key)
        d = self.dictionary
        for k in key:
            if not isinstance(d, dict):
                raise RepositoryKeyError(key)
            try:
                d = d[k]
            except KeyError:
                with self.lock:
                    return self.repository.list(key)
        if not isinstance(d, dict):
            raise RepositoryKeyError(key)
        try:
            with self.lock:
                src = self.repository.list(key)
        except RepositoryKeyError:
            return d
        return frozenset(d).union(src)

    def flush(self, _dictionary=None, _key=None):
        """Flush all buffered updates to the :attr:`repository`."""
        with self.lock if _dictionary is None else self.dump_context():
            if _dictionary is None:
                _dictionary = self.dictionary
                _key = ()
            items = getattr(_dictionary, 'iteritems', _dictionary.items)()
            read_from_repository = self.repository.read
            write_to_repository = self.repository.write
            for key, value in items:
                key = _key + (key,)
                if isinstance(value, dict):
                    self.flush(_dictionary=value, _key=key)
                else:
                    type_hint, bytearray = value
                    bytearray = bytearray,
                    if type_hint is not None:
                        try:
                            prev_iterable = read_from_repository(key)
                        except RepositoryKeyError:
                            pass
                        else:
                            prev_iterable = list(prev_iterable)
                            prev = parse_revision(prev_iterable)
                            crev = parse_revision(bytearray)
                            if prev is not None and \
                                (crev is None or crev[0] is None or
                                 not crev[1].contains(prev[0])):
                                prev_doc = read(type_hint, prev_iterable)
                                doc = read(type_hint, bytearray)
                                merged_doc = prev[0].session.merge(
                                    doc,
                                    prev_doc,
                                    force=True
                                )
                                bytearray = write(
                                    merged_doc,
                                    canonical_order=True,
                                    as_bytes=True
                                )
                    write_to_repository(key, bytearray)
            _dictionary.clear()

    @contextlib.contextmanager
    def dump_context(self):
        yield

    def __repr__(self):
        return '{0.__module__}.{0.__name__}({1!r})'.format(type(self),
                                                           self.repository)


class TransactionError(RuntimeError):
    """The error that rises if there's no ongoing transaction while it's
    needed to update the stage, or if there's already begun ongoing transaction
    when the new transaction get tried to begin.

    """


class Route(object):
    """Descriptor that routes a ``document_type`` to a particular key
    path pattern in the repository.

    ``key_spec`` could contain some format strings.  Format strings can
    take a keyword (``session``) and zero or more positional arguments.

    For example, if you route a document type without any positional
    arguments in ``key_spec`` format::

        class Stage(BaseStage):
            '''Stage example.'''

            metadata = Route(
                Metadata,
                ['metadata', '{session.identifier}.xml']
            )

    Stage instance will has a ``metadata`` attribute that simply holds
    ``Metadata`` document instance (in the example):

    >>> stage.metadata  # ['metadata', 'session-id.xml']
    <Metadata ...>

    If you route something with one or more positional arguments in
    ``key_spec`` format, then it works in some different way::

        class Stage(BaseStage):
            '''Stage example.'''

            seating_chart = Route(
                Student,
                ['students', 'col-{0}', 'row-{1}', '{session.identifier}.xml']
            )

    In the above routing, two positional arguments were used.  It means that
    the ``seating_chart`` property will return two-dimensional mapping object
    (:class:`Directory`):

    >>> stage.seating_chart  # ['students', ...]
    <libearth.directory.Directory ['students']>
    >>> list(stage.seating_chart)
    ['A', 'B', 'C', 'D']
    >>> b = stage.seating_chart['B']  # ['students', 'col-B', ...]
    <libearth.directory.Directory ['students', 'col-B']>
    >>> list(stage.seating_chart['B'])
    ['1', '2', '3', '4', '5', '6']
    >>> stage.seating_chart['B']['6']  \\
    ... # ['students', 'col-B', 'row-6', 'session-id.xml']
    <Student B6>

    :param document_type: the type of document to route.
                          it has to be a subclass of
                          :class:`~libearth.session.MergeableDocumentElement`
    :type document_type: :class:`type`
    :param key_spec: the repository key pattern that might contain some
                     format strings
                     e.g. ``['docs', '{0}', '{session.identifier}.xml']`.
                     positional values are used for directory indices
                     (if present), and ``session`` keyword value is used
                     for identifying sessions
    :type key_spec: :class:`collections.Sequence`

    """

    #: (:class:`type`) The type of the routed document.  It is a subtype of
    #: :class:`~libearth.session.MergeableDocumentElement`.
    document_type = None

    #: (:class:`collections.Sequence`) The repository key pattern that
    #: might contain some format strings.
    key_spec = None

    def __init__(self, document_type, key_spec):
        if not isinstance(document_type, type):
            raise TypeError('document_type must be a type object, '
                            'not {1!r}'.format(document_type))
        elif not issubclass(document_type, MergeableDocumentElement):
            raise TypeError(
                'document_type must be a subtype of {0.__module__}.'
                '{0.__name__}, not {1.__module__}.{1.__name__}'.format(
                    MergeableDocumentElement,
                    document_type
                )
            )
        elif not isinstance(key_spec, collections.Sequence):
            raise TypeError('key_spec must be a sequence, not ' +
                            repr(key_spec))
        self.document_type = document_type
        self.key_spec = key_spec

    def __get__(self, obj, cls=None):
        if obj is None or isinstance(obj, type):
            return self
        # Should be merged!
        assert isinstance(obj, BaseStage)
        key = []
        for fmt in self.key_spec:
            try:
                chunk = fmt.format(session=obj.session)
            except IndexError:
                return Directory(obj, self.document_type,
                                 self.key_spec, (), key)
            try:
                fmt.format()
            except KeyError:
                break
            else:
                key.append(chunk)
        return obj.read_merged_document(self.document_type, self.key_spec, key)

    def __set__(self, obj, value):
        session = obj.session
        try:
            key = [fmt.format(session=session) for fmt in self.key_spec]
        except IndexError:
            raise AttributeError('cannot set the directory')
        obj.write(key, value)


def compile_format_to_pattern(format_string):
    """Compile a ``format_string`` to regular expression pattern.
    For example, ``'string{0}like{1}this{{2}}'`` will be compiled to
    ``/^string(.*?)like(.*?)this\{2\}$/``.

    :param format_string: format string to compile
    :type format_string: :class:`str`
    :returns: compiled pattern object
    :rtype: :class:`re.RegexObject`

    """
    pattern = ['^']
    i = 0
    for match in re.finditer(r'(^|[^{])\{[^}]+\}|(\{\{)|(\}\})', format_string):
        if match.group(2):
            j = match.start()
            chunk = r'\{'
        elif match.group(3):
            j = match.start()
            chunk = r'\}'
        else:
            j = match.end(1)
            chunk = '(.*?)'
        pattern.append(re.escape(format_string[i:j]))
        pattern.append(chunk)
        i = match.end(0)
    if len(format_string) > i:
        pattern.append(re.escape(format_string[i:]))
    pattern.append('$')
    return re.compile(''.join(pattern))


class Directory(collections.Mapping):
    """Mapping object which represents hierarchy of routed key path.

    :param stage: the current stage
    :type stage: :class:`BaseStage`
    :param document_type: the same to :attr:`Route.document_type`
    :type document_type: :class:`type`
    :param key_spec: the same to :attr:`Route.key_spec` value
    :type key_spec: :class:`collections.Sequence`
    :param indices: the upper indices that are already completed
    :type indices: :class:`collections.Sequence`
    :param key: the upper key that are already completed
    :type key: :class:`collections.Sequence`

    .. note::

       The constructor is intended to be internal, so don't instantiate
       it directory.  Use :class:`Route` instead.

    """

    def __init__(self, stage, document_type, key_spec, indices, key):
        if not isinstance(stage, BaseStage):
            raise TypeError('stage must be an instance of {0.__module__}.'
                            '{0.__name__}, not {1!r}'.format(BaseStage, stage))
        elif not isinstance(document_type, type):
            raise TypeError('document_type must be a type object, '
                            'not {1!r}'.format(document_type))
        elif not issubclass(document_type, MergeableDocumentElement):
            raise TypeError(
                'document_type must be a subtype of {0.__module__}.'
                '{0.__name__}, not {1.__module__}.{1.__name__}'.format(
                    MergeableDocumentElement,
                    document_type
                )
            )
        elif not isinstance(key_spec, collections.Sequence):
            raise TypeError('key_spec must be a sequence, not ' +
                            repr(key_spec))
        elif not isinstance(indices, collections.Sequence):
            raise TypeError('indices must be a sequence, not ' + repr(indices))
        elif not isinstance(key, collections.Sequence):
            raise TypeError('key must be a sequence, not ' + repr(key))
        elif len(key) >= len(key_spec):
            raise ValueError('key seems already complete')
        self.stage = stage
        self.document_type = document_type
        self.key_spec = key_spec
        self.indices = tuple(indices)
        self.key = key

    def __len__(self):
        return sum(1 for _ in self)

    def __getitem__(self, index):
        key = list(self.key)
        indices = self.indices + (index,)
        stage = self.stage
        session = stage.session
        for fmt in self.key_spec[len(key):]:
            try:
                chunk = fmt.format(*indices, session=session)
            except IndexError:
                if stage.repository.exists(key):
                    return Directory(stage, self.document_type,
                                     self.key_spec, indices, key)
                raise KeyError(index)
            try:
                fmt.format(*indices)
            except KeyError:
                break
            else:
                key.append(chunk)
        try:
            doc = stage.read_merged_document(self.document_type,
                                             self.key_spec,
                                             key)
        except RepositoryKeyError:
            raise KeyError(index)
        if doc:
            return doc
        raise KeyError(index)

    def __setitem__(self, index, doc):
        indices = self.indices + (index,)
        session = self.stage.session
        try:
            key = [fmt.format(*indices, session=session)
                   for fmt in self.key_spec]
        except IndexError:
            raise TypeError('cannot set the directory')
        self.stage.write(key, doc)

    def __iter__(self):
        it = self.stage.repository.list(self.key)
        pattern = compile_format_to_pattern(self.key_spec[len(self.key)])
        indices = set()
        for key in it:
            match = pattern.match(key)
            if match:
                index = match.group(1)
                if index not in indices:
                    yield index
                    indices.add(index)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} {1!r}>'.format(
            type(self), self.key
        )


class Stage(BaseStage):
    """Staged documents of Earth Reader."""

    #: (:class:`collections.MutableMapping`) The map of feed ids to
    #: :class:`~libearth.feed.Feed` objects.
    feeds = Route(Feed, ['feeds', '{0}', '{session.identifier}.xml'])

    #: (:class:`~libearth.subscribe.SubscriptionList`) The set of subscriptions.
    subscriptions = Route(SubscriptionList,
                          ['subscriptions.{session.identifier}.xml'])



import os.path
import sys

try:
    from setuptools import find_packages, setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import find_packages, setup

from libearth.version import VERSION


def readme():
    try:
        with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
            return f.read()
    except (IOError, OSError):
        return ''


install_requires = []
if sys.version_info < (3, 2):
    install_requires.append('futures')


setup(
    name='libearth',
    version=VERSION,
    description='The shared common library for Earth Reader apps',
    long_description=readme(),
    url='http://libearth.earthreader.org/',
    download_url='https://github.com/earthreader/libearth/releases',
    author='Hong Minhee',
    author_email='minhee' '@' 'dahlia.kr',
    license='GPLv2 or later',
    packages=find_packages(exclude=['tests']),
    entry_points='''
        [libearth.repositories]
        file = libearth.repository:FileSystemRepository
    ''',
    install_requires=install_requires,
    tests_require=['pytest >= 2.4.0'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved ::'
        ' GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: IronPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Communications',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: XML'
    ]
)



""":mod:`libearth.subscribe` --- Subscription list
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Maintain the subscription list using OPML_ format, which is de facto standard
for the purpose.

.. _OPML: http://dev.opml.org/spec2.html

"""
import collections
import datetime
import distutils.version
import hashlib

from .codecs import Boolean, Integer, Rfc822
from .compat import string_type, text_type
from .feed import Feed, Person
from .schema import Attribute, Child, Codec, Element, Text
from .session import MergeableDocumentElement
from .tz import now

__all__ = ('Body', 'Category', 'CommaSeparatedList', 'Head', 'Outline',
           'Subscription', 'SubscriptionList', 'SubscriptionSet')


#: (:class:`str`) The XML namespace name used for Earth Reader subscription
#: list metadata.
METADATA_XMLNS = 'http://earthreader.org/subscription-list/'


class CommaSeparatedList(Codec):
    """Encode strings e.g. ``['a', 'b', 'c']`` into a comma-separated list
    e.g. ``'a,b,c'``, and decode it back to a Python list.  Whitespaces
    between commas are ignored.

    >>> codec = CommaSeparatedList()
    >>> codec.encode(['technology', 'business'])
    'technology,business'
    >>> codec.decode('technology, business')
    ['technology', 'business']

    """

    def encode(self, value):
        if value is None:
            res = ""
        elif isinstance(value, text_type):
            res = value
        else:
            res = ",".join(value)
        return res

    def decode(self, text):
        if text is None:
            lst = []
        else:
            lst = [elem.strip() for elem in text.split(',')]
        return lst


class SubscriptionSet(collections.MutableSet):
    """Mixin for :class:`SubscriptionList` and :class:`Category`, both can
    group :class:`Subscription` object and other :class:`Category` objects,
    to implement :class:`collections.MutableSet` protocol.

    """

    @property
    def children(self):
        """(:class:`collections.MutableSequence`) Child :class:`Outline`
         objects.

        .. note::

           Every subclass of :class:`SubscriptionSet` has to override
           :attr:`children` property to implement details.

         """
        raise NotImplementedError(
            'every subclass of {0.__module__}.{0.__name__} has to '
            'implement children property'.format(SubscriptionSet)
        )

    def __len__(self):
        if not self.children:
            return 0
        return sum(not child.deleted for child in self.children)

    def __iter__(self):
        categories = set()
        subscriptions = set()
        for outline in self.children:
            if outline.deleted:
                continue
            elif (outline.type == 'rss' or outline.feed_uri or
                  isinstance(outline, Subscription)):
                if outline.feed_uri in subscriptions:
                    continue
                subscriptions.add(outline.feed_uri)
                if not isinstance(outline, Subscription):
                    if not outline.feed_id:
                        outline.feed_id = hashlib.sha1(
                            outline.feed_uri.encode('utf-8')
                        ).hexdigest()
                    outline.__class__ = Subscription
                    outline._title = outline.label
                yield outline
            elif outline.label in categories:
                continue
            else:
                categories.add(outline.label)
                if not isinstance(outline, Category):
                    outline.__class__ = Category
                    outline._title = outline.label
                yield outline

    def contains(self, outline, recursively=False):
        """Determine whether the set contains the given ``outline``.
        If ``recursively`` is :const:`False` (which is by default)
        it works in the same way to :keyword:`in` operator.

        :param outline: the subscription or category to find
        :type outline: :class:`Outline`
        :param recursively: if it's :const:`True` find the ``outline``
                            in the whole tree, or if :const:`False` find
                            it in only its direct children.
                            :const:`False` by default
        :type recursively: :class:`bool`
        :returns: :const:`True` if the set (or tree) contains the given
                  ``outline``, or :const:`False`
        :rtype: :class:`bool`

        .. versionadded:: 0.2.0

        """
        if not isinstance(outline, Outline):
            raise TypeError('expected an instance of {0.__module__}.'
                            '{0.__name__}, not {1!r}'.format(Outline, outline))
        for child in self.children:
            if outline == child:
                if recursively:
                    return not child.deleted
                elif not child.deleted:
                    return True
        if recursively:
            for subcategory in self:
                if isinstance(subcategory, SubscriptionSet) and \
                   subcategory.contains(outline, recursively=True):
                    return True
        return False

    def __contains__(self, outline):
        try:
            return self.contains(outline, recursively=False)
        except TypeError:
            return False

    def add(self, value):
        if not isinstance(value, Outline):
            raise TypeError('expected {0.__module__}.{0.__name__}, not '
                            '{1!r}'.format(Outline, value))
        if value.type == 'rss' or isinstance(value, Subscription):
            value.type = 'rss'
            for outline in self.children:
                if not (outline.type == 'rss' or
                        isinstance(outline, Subscription)):
                    continue
                if outline.feed_uri == value.feed_uri:
                    outline.created_at = now()
                    return
        else:
            value.type = 'category'
            for outline in self.children:
                if not (outline.type == 'category' or
                        isinstance(outline, Category)):
                    continue
                if outline.label == value.label:
                    outline.created_at = now()
                    return
        value.created_at = now()
        self.children.append(value)

    def discard(self, outline):
        if not isinstance(outline, Outline):
            raise TypeError('expected {0.__module__}.{0.__name__}, not '
                            '{1!r}'.format(Outline, outline))
        deleted_at = now()
        if deleted_at <= outline.created_at:
            deleted_at = (outline.created_at +
                          datetime.timedelta(microseconds=1))  # FIXME
        for child in self.children:
            if child == outline:
                child.deleted_at = deleted_at
                assert child.deleted

    def subscribe(self, feed, icon_uri=None):
        """Add a subscription from :class:`~libearth.feed.Feed` instance.
        Prefer this method over :meth:`add()` method.

        :param feed: feed to subscribe
        :type feed: :class:`~libearth.feed.Feed`
        :param icon_uri: optional favicon url of the ``feed``
        :type icon_uri: :class:`str`
        :returns: the created subscription object
        :rtype: :class:`Subscription`

        .. versionadded:: 0.3.0
           Optional ``icon_url`` parameter was added.

        """
        if not isinstance(feed, Feed):
            raise TypeError('feed must be an instance of {0.__module__}.'
                            '{0.__name__}, not {1!r}'.format(Feed, feed))
        elif icon_uri is None:
            favicon = feed.links.favicon
            if favicon is not None:
                icon_uri = favicon.uri
        elif not isinstance(icon_uri, string_type):
            raise TypeError('icon_uri must be a string, not ' +
                            repr(icon_uri))
        sub = Subscription(
            feed_id=hashlib.sha1(feed.id.encode('utf-8')).hexdigest(),
            icon_uri=icon_uri,
            label=text_type(feed.title),
            _title=text_type(feed.title),
            feed_uri=next(l.uri for l in feed.links if l.relation == 'self'),
            alternate_uri=next(
                (l.uri for l in feed.links
                 if l.relation == 'alternate' and l.mimetype == 'text/html'),
                None
            ),
            created_at=now()
        )
        for child in self.children:
            if child == sub:
                self.children.remove(child)
        self.children.append(sub)
        return sub

    @property
    def categories(self):
        """(:class:`collections.Mapping`) Label to :class:`Category` instance
        mapping.

        """
        categories = {}
        for child in self:
            if isinstance(child, Category):
                categories[child.label] = child
        return categories

    @property
    def subscriptions(self):
        """(:class:`collections.Set`) The subset which consists of only
        :class:`Subscription` instances.

        """
        return frozenset(e for e in self if isinstance(e, Subscription))

    @property
    def recursive_subscriptions(self):
        subscriptions = set()
        for child in self:
            if isinstance(child, Subscription):
                subscriptions.add(child)
            elif isinstance(child, Category):
                subscriptions.update(child.recursive_subscriptions)
        return subscriptions

    def __merge_entities__(self, other):
        for outline in other.children:
            for child in self.children:
                if child == outline:
                    if type(child) is Outline:
                        child.__class__ = (Subscription
                                           if child.type == 'rss'
                                           else Category)
                        outline._title = outline.label
                    child.created_at = max(child.created_at,
                                           outline.created_at)
                    if not child.deleted_at:
                        child.deleted_at = outline.deleted_at
                    elif outline.deleted_at:
                        child.deleted_at = max(child.deleted_at,
                                               outline.deleted_at)
                    if child.children and child.deleted:
                        del child.children[:]
                    if child.children or outline.children:
                        SubscriptionSet.__merge_entities__(child, outline)
                    break
            else:
                self.add(outline)
        return self


class Outline(Element):
    """Represent ``outline`` element of OPML document."""

    #: (:class:`str`) The human-readable text of the outline.
    label = Attribute('text', required=True)

    #: (:class:`str`) Internally-used type identifier.
    type = Attribute('type')

    #: (:class:`datetime.datetime`) The created time.
    created_at = Attribute('created', Rfc822(microseconds=True))

    #: (:class:`datetime.datetime`) The archived time, if deleted ever.
    #: It could be :const:`None` as well if it's never deleted.
    #: Note that it doesn't have enough information about whether
    #: it's actually deleted or not.  For that you have to use
    #: :attr:`deleted` property instead.
    #:
    #: .. versionadded:: 0.3.0
    deleted_at = Attribute(
        'deleted',
        Rfc822(microseconds=True),
        xmlns=METADATA_XMLNS
    )

    feed_uri = Attribute('xmlUrl')
    alternate_uri = Attribute('htmlUrl')
    children = Child('outline', 'Outline', multiple=True)
    feed_id = Attribute('id', xmlns=METADATA_XMLNS)

    _title = Attribute('title')
    _category = Attribute('category', CommaSeparatedList)
    _breakpoint = Attribute('isBreakpoint', Boolean)

    @property
    def deleted(self):
        """(:class:`bool`) Whether it is deleted (archived) or not.

        .. versionadded:: 0.3.0

        """
        return bool(self.deleted_at and self.deleted_at > self.created_at)

    def __eq__(self, other):
        if isinstance(other, Outline):
            if self.type == 'rss':
                return self.feed_uri == other.feed_uri
            return self.label == other.label
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        if self.type == 'rss':
            return hash(self.feed_uri)
        return hash(self.label)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} type={1} label={2}{3}>'.format(
            type(self), repr(self.type), repr(self.label),
            ' [deleted]' if self.deleted else ''
        )


class Category(Outline, SubscriptionSet):
    """Category which groups :class:`Subscription` objects or other
    :class:`Category` objects.  It implements :class:`collections.MutableSet`
    protocol.

    .. attribute:: children

       (:class:`collections.MutableSequence`) The list of child :class:`Outline`
       elements.  It's for internal use.

    """

    type = Attribute('type', default=lambda _: 'category')

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} {1!r}{2}>'.format(
            type(self), self.label, ' [deleted]' if self.deleted else ''
        )


class Subscription(Outline):
    """Subscription which holds referring :attr:`feed_uri`.

    .. attribute:: feed_id

       (:class:`str`) The feed identifier to be used for lookup.
       It's intended to be SHA1 digest of :class:`Feed.id
       <libearth.feed.Feed.id>` value (which is UTF-8 encoded).

    .. attribute:: feed_uri

       (:class:`str`) The feed url.

    .. attribute:: alternate_uri

       (:class:`str`) The web page url.

    """

    type = Attribute('type', default=lambda _: 'rss')

    #: (:class:`str`) Optional favicon url.
    #:
    #: .. versionadded:: 0.3.0
    icon_uri = Attribute('icon', xmlns=METADATA_XMLNS, default=lambda _: None)

    def __repr__(self):
        return '<{0.__module__}.{0.__name__} {1} {2!r} ({3!r}){4}>'.format(
            type(self), self.feed_id, self.label, self.feed_uri,
            ' [deleted]' if self.deleted else ''
        )


class Head(Element):
    """Represent ``head`` element of OPML document."""

    #: (:class:`str`) The title of the subscription list.
    title = Text('title')

    #: (:class:`str`) The owner's name.
    owner_name = Text('ownerName')

    #: (:class:`str`) The owner's email.
    owner_email = Text('ownerEmail')

    #: (:class:`str`) The owner's website url.
    owner_uri = Text('ownerId')

    created_at = Text('dateCreated', Rfc822)
    updated_at = Text('dateModified', Rfc822)

    _docs = Text('docs')
    _expansion_state = Text('expansionState', CommaSeparatedList)
    _vert_scroll_state = Text('vertScrollState', Integer)
    _window_top = Text('windowTop', Integer)
    _window_bottom = Text('windowBottom', Integer)
    _window_left = Text('windowLeft', Integer)
    _window_right = Text('windowRight', Integer)


class Body(Element):
    """Represent ``body`` element of OPML document."""

    #: (:class:`collections.MutableSequence`) Child :class:`Outline` objects.
    children = Child('outline', Outline, multiple=True)


class SubscriptionList(MergeableDocumentElement, SubscriptionSet):
    """The set (exactly, tree) of subscriptions.  It consists of
    :class:`Subscription`\ s and :class:`Category` objects for grouping.
    It implements :class:`collections.MutableSet` protocol.

    """

    __tag__ = 'opml'
    head = Child('head', Head)
    body = Child('body', Body)

    #: (:class:`distutils.version.StrictVersion`) The OPML version number.
    version = Attribute(
        'version',
        encoder=str,
        decoder=distutils.version.StrictVersion,
        default=lambda _: distutils.version.StrictVersion('2.0')
    )

    @property
    def children(self):
        if self.body is None:
            self.body = Body()
        return self.body.children

    @property
    def title(self):
        """(:class:`str`) The title of the subscription list."""
        head = self.head
        return head and head.title

    @title.setter
    def title(self, title):
        head = self.head
        if head is None:
            head = Head()
            self.head = head
        head.title = title

    @property
    def owner(self):
        """(:class:`~libearth.feed.Person`) The owner of the subscription
        list.

        """
        head = self.head
        if head is None:
            return
        if head.owner_name is None and head.owner_email is None and \
           head.owner_uri is None:
            return
        return Person(
            name=head.owner_name,
            email=head.owner_email,
            uri=head.owner_uri
        )

    @owner.setter
    def owner(self, owner):
        head = self.head
        if head is None:
            if owner is None:
                return
            head = self.head = Head()
        elif owner is None:
            head.owner_name = head.owner_email = head.owner_uri = None
            return
        head.owner_name = owner.name
        head.owner_email = owner.email
        head.owner_uri = owner.uri

    def __merge_entities__(self, other):
        subs = SubscriptionSet.__merge_entities__(self, other)
        doc = MergeableDocumentElement.__merge_entities__(self, other)
        doc.body = subs.body
        return doc

    def __repr__(self):
        head = self.head
        if head is None:
            return super(SubscriptionList, self).__repr__()
        return '<{0.__module__}.{0.__name__} {1!r} of {2} <{3}>>'.format(
            type(self), head.title, head.owner_name,
            head.owner_email or head.owner_uri
        )
